#ifndef RTOS_H
#define RTOS_H

#include <stdint.h>

#define true   -1
#define false  0
#define TRUE   -1
#define FALSE  0

// ---------- types to use in the using program ------------

typedef uint8_t     BOOL;
typedef uint8_t     BYTE;
typedef uint16_t    WORD;
typedef uint32_t    DWORD;

typedef BYTE        TOS_Task;
typedef BYTE        TOS_State;

#ifdef OS_SEMAPHORES
typedef BYTE        TOS_BinarySemaphore;
typedef BYTE        TOS_CountingSemaphore;
typedef BYTE        TOS_Semaphore;
#endif

#ifdef OS_PRIORITIES
typedef BYTE        TOS_Priority;
#endif
// ------------------------------------------------

// ---------- constants to use in the using program --------

#define OS_NO_TASK   255 // no free task, (OS_CreateTask result)
#ifdef OS_SEMAPHORES
   #define OS_NO_EVENT  255 // no free event, (OS_CreateXxxSemaphore result)
   #ifdef OS_TIMEOUT
      #define OS_NO_TIMEOUT  0 // waiting for an event with no timeout (parameter for "OS_Wait..." procedures)
   #endif
#endif

// TOS_State type: possible values (return value of "OS_State")
// task states (used in the TaskData)
#define st_DESTROYED  0 // task not created
#define st_STOPPED    1 // Task created but not started (or has been stopped)
#define st_DELAYED    2 // Task has been started but is suspended for a period (waiting for n ticks)
#define st_WAITING    3 // Task has been started and is waiting an event forever or with time-out
#define st_ELIGIBLE   4 // Task has been started and is eligible to run
#define st_RUNNING    5 // Task is the currently active (executing) Task


// ----------------------- published routines -----------------

// ----------------
// General routines
// ----------------

void OS_Init();
// To be called before any other RTOS routine can be called.

void OS_Run();
// Starts the RTOS scheduler.
// Should be the last routine executed in the main program, the task is blocking (never left).

extern void InitTimer1(); // to be delivered by the user

typedef void (*TOS_TaskProcPtr)();
typedef void (*TProcPtr)();

#ifdef OS_IDLE
TProcPtr OS_IdleProcedure;   // void (*OS_IdleProcedure)();
#endif

// ---------------------
// Task related routines
// ---------------------

#ifdef OS_PRIORITIES
TOS_Task OS_CreateTask(TOS_TaskProcPtr TaskProc, TOS_Priority Priority);
#else
TOS_Task OS_CreateTask(TOS_TaskProcPtr TaskProc);
#endif
// Adds a task to the TasksList. The "Task" parameter is the startaddress of the task.
// The function returns the number of the task in the taskslist or "OS_NO_TASK" when
// the taskslist is full.

void OS_StartTask(TOS_Task TaskID);
// Starts a (stopped) task identified by "TaskID"

#ifdef OS_DELAY
void OS_StartTask_Delay(TOS_Task TaskID, WORD Ticks);
// Starts a (stopped) task identified by "TaskID" after an initial waiting time of "Ticks" timebase ticks
#endif

void OS_StopTask(TOS_Task TaskID);
// Stops the task "TaskID" and returns to the scheduler if "TaskID" is the current task.
// The task can only be restarted with "OS_StartTask" or "OS_StartTask_Delay" which will make the task resume where it was when stopped.

void OS_ReplaceTask(TOS_Task TaskID);
// Stops the current task and starts the task with "TaskID"

TOS_Task OS_CurrentTask();
// Returns the ID of the currently active task

BOOL OS_TaskRunning(TOS_Task TaskID);
// Returns true if the task "TaskID" is started, otherwise false

TOS_State OS_TaskState(TOS_Task TaskID);
// Returns the state of "TaskID"


// -------------------------------
// Task switching related routines
// -------------------------------

void OS_Yield();
// Unconditionally yields to the scheduler. If no other task is waiting to run then the
// current task will resume at the next instruction after "OS_Yield".

#ifdef OS_DELAY
void OS_Delay(WORD Ticks);
// Suspends the current task and returns to the scheduler. The latter will resume the task
// after "Ticks" time
#endif

// --------------------------
// Semaphore related routines
// --------------------------

#ifdef OS_SEMAPHORES
TOS_BinarySemaphore OS_CreateBinarySemaphore(BOOL InitialValue);
// Adds a binary semaphore to the eventslist.
// The function returns the semaphore's number or "OS_NO_EVENT" if the eventslist is full
// "InitialValue" is the initial value assigned to the semaphore (true = "signalled").

TOS_CountingSemaphore OS_CreateCountingSemaphore(WORD InitialValue);
// Adds a counting semaphore to the eventslist.
// The function returns the semaphore's number or "OS_NO_EVENT" if the eventslist is full
// "InitialValue" is the initial count value assigned to the semaphore.

void OS_SignalSemaphore(TOS_Semaphore Event_);
// Sets the semaphore to "signalled" (binary semaphore) or increments the semaphore count by one (counting semaphore)

void OS_ClearSemaphore(TOS_Semaphore Event_);
// Sets the semaphore to "not signalled" (binary semaphore) or sets the semaphore count to zero (counting semaphore).
// Also pending signaling from interrupts are cleared.

#ifdef OS_SIGNAL_ISR
void OS_SignalSemaphore_ISR(TOS_Semaphore Event_);
// To be used from and only from within interrupt service routines.
// Sets the semaphore to "signalled" (binary semaphore) or increments the semaphore count by one
// (counting semaphore)
#endif

#ifdef OS_TIMEOUT
void OS_WaitSemaphore(TOS_Semaphore Event_, WORD TimeOut);
#else
void OS_WaitSemaphore(TOS_Semaphore Event_);
#endif
// Suspends the current task until the binary or counting semaphore referenced in "Event_"
// has been signalled (binary semaphore) or the semaphore count has a value > 0 (counting semaphore)
// or the TimeOut (ticks) has been elapsed.

// If the event is already signalled (binary) or the semaphore count has a value > 0 (counting) when the
// wait routine is called then the current task will continue (no task switching).

// After a "OS_WaitSemaphore" the semaphore will be cleared (binary) or the semaphore count value will
// be decremented by one (counting semaphore, provided the value is > 0).

// If the wait times out the task will be resumed with the "OS_TimeOut" flag set. If the
// event is signalled within the TimeOut clicks the task will be resumed with the
// "OS_TimeOut" flag cleared. Use "OS_NO_TIMEOUT" as "timeOut" parameter when an infinite timeout
// is needed.

BOOL OS_ReadSemaphore(TOS_Semaphore Event_);
// Returns true if the binary semaphore signalled or the counting semaphore's value is > 0, otherwise false
// The value of the semaphore is not changed (signalling or count value)

WORD OS_ReadCountingSemaphore(TOS_Semaphore Event_);
// Returns the counting semaphore's value.
// The value of the semaphore is not changed.

BOOL OS_TrySemaphore(TOS_Semaphore Event_);
// Returns true if the binary semaphore is signalled or the counting semaphore's value is > 0,
// otherwise false.
// The value of the semaphore remains unchanged.

#ifdef OS_TIMEOUT
BOOL OS_TimeOut();
// Returns "true" if a timeout occured, or false if no timeout occurred (e.g. after "OS_WaitSemaphore").
// After calling "OS_TimeOut" the timeout flag is cleared.
// The flag is also initially cleared when an "OS_Wait..." routine is called.
#endif
#endif

// ---------------------------------
// Critical section related routines
// ---------------------------------

void OS_EnterCriticalSection();
// will prevent task switching (until a call to OS_LeaveCriticalSection)

void OS_LeaveCriticalSection();
// will allow task switching again after a call to OS_EnterCriticalSection

// -------------------------
// Message related routines
// -------------------------

// not implemented yet

// -------------------------
// Flags related routines
// -------------------------

// not implemented yet

// -------------------------
// Priority related routines
// -------------------------

#ifdef OS_PRIORITIES
void OS_SetPriority(TOS_Task TaskID, TOS_Priority Priority);
// Sets the priority of "TaskID" to "Priority"

TOS_Priority OS_Priority(TOS_Task TaskID);
// Returns the priority of "TaskID"
#endif

// -------------------------
// Stack check related routines
// ------------------------

#ifdef OS_STACK_CHECK
BYTE OS_Stack_Usage(TOS_Task TaskID);
// Returns the usage of the current task stack in percent
#endif

#endif