// D. Rosseel
// Original: 2015-07-28
// Latest update: 2015-08-09

/*
History
2015-07-28: Initial version.
2015-08-06: - Added the "Critical Section" procedures.
            - R1 is not part of the context any more (R1=GP never changes).
            - Made the Timer1Int routine does not make any "calls" anymore.
            - Cleaned up code and added comments.
2015-08-09: - Added OS_NO_AUTO compiler directive (in RTOS_Defines.pld) to disable automatic task switching,
            setting this directive makes the system a cooperative RTOS (CRTOS).
            - Always restart the 10 timerticks counter when a new task is (to be) selected.
            - Forced the first task selection now (in OS_Run): faster starting of first task.
            - Added the "Check_Events_ISR" again in the semaphore test/read routines to prevent 1 timertick delay.
            - Made "RestoreInterrupts" inline code. In some cases a call to a routine is not allowed.
2015-08-14: - Added the instructions "DI" and "EHB" to the epilogue code of the OS interrupt routine.
            - Fetched the interrupt Priority from the "Cause" coprocessor register to put in the "Status" register
            - Added always a "rdpgpr sp, sp" to the beginning of the interrupt prologue
*/


// ------------------------- INTERFACE -----------------------------------------

// ----------- RTOS Configuration ----------------------------------------------

// RTOS capabilities

#include "RTOS_Config.h"
#include <stddef.h>

#ifdef OS_DELAY
#define OS_DELAY_TIMEOUT
#endif

#ifdef OS_TIMEOUT
#define OS_DELAY_TIMEOUT
#endif

#ifdef OS_IDLE
TProcPtr OS_IdleProcedure;   // void (*OS_IdleProcedure)();
#endif

#define IntTime  10   // Task switch interrupt period (in millisecs)

sbit OS_Timer_IE_bit at T1IE_bit;


#ifdef OS_SEMAPHORES
//Event types (used in the EventData)
#define et_NOEVENT    0
#define et_BINARY     2  // binary semaphore type
#define et_COUNTING   3  // counting semaphore type
//#define et_MESSAGE    3  // Message type
//#define et_EVENT      4  // event type
#endif

typedef DWORD TRegs[30];

typedef struct {
    // ------- task context -------------
    TRegs Regs;
    DWORD RetAddr,     // saved interrupt return address
    HiVal,       // saved value of the HI register
    LoVal,       // saved value of the LO register
    Status_,     // saved value of coprocessor 0 status
    SRSCtl;
    // ----------------------------------

    BYTE _State;    // see above const definition "task states"
    #ifdef OS_PRIORITIES
    TOS_Priority _Priority; // priorities will range from 0 (highest) to OS_PRIORITY_COUNT - 1 (lowest)
    #endif
    #ifdef OS_DELAY_TIMEOUT
    WORD _Timer;    // timer/counter for delays, timeout etc
    #endif
    #ifdef OS_SEMAPHORES
    BYTE _EventID;    // Event ID waiting for
    #endif
    BYTE _Misc;    // see bit assignments below
    #ifdef OS_STACK_CHECK
    DWORD _StackBase; // startaddress of the stack (lowest address);
    #endif
} TTaskData;

#ifdef OS_TIMEOUT
// bits in TTaskData._Misc:
#define mi_TIMEOUT_bit     B0  // bit 0 = timeout while waiting for an event (e.g. semaphore)
#endif

#ifdef OS_SEMAPHORES
typedef struct {
    BYTE _Type; // see definition above
    WORD _SemCount; // semaphore count (for binary semaphores: only 0 and 1 valid )
    #ifdef OS_SIGNAL_ISR
    WORD _SemISR; // temporary storage for signaling from ISR
    #endif
} TEventData;
#endif

TTaskData TaskData[OS_TASKS_COUNT];
TTaskData IdleData;
#ifdef OS_SEMAPHORES
TEventData EventData[OS_EVENTS_COUNT];
#endif
#ifdef OS_PRIORITIES
TOS_Task LastExecutedTask[OS_PRIORITY_COUNT];
BYTE LowestPriority;
#else
TOS_Task LastExecutedTask;
#endif

BOOL OSRunning;
BYTE CurrentTask, // current task executing
LastTask,    // task with the highest ID (last created)
LastEvent;   // event with the highest ID (last created)

BOOL IdleRunning,
Immediate;

TTaskData *Ptr;
DWORD InterruptState;

BYTE OS_Init_I;

void OS_Init() {
    OS_Timer_IE_bit = 0; // disable the timebase timer interrupts

    // disable the timer1 interrupt also by setting its Priority to 0
    T1IP0_bit     = 0;
    T1IP1_bit     = 0;
    T1IP2_bit     = 0;

    T1IF_bit      = 0;

    for (OS_Init_I = 0; OS_Init_I < OS_TASKS_COUNT; OS_Init_I++)
    TaskData[OS_Init_I]._State = st_DESTROYED;  // remove all tasks from the task list

#ifdef OS_SEMAPHORES
    for (OS_Init_I = 0; OS_Init_I < OS_EVENTS_COUNT; OS_Init_I++)
    EventData[OS_Init_I]._Type = et_NOEVENT;    // remove all events from the event list
    LastEvent = OS_NO_EVENT;
#endif

#ifdef OS_PRIORITIES
    for (OS_Init_I = 0; OS_Init_I < OS_PRIORITY_COUNT; OS_Init_I++)
    LastExecutedTask[OS_Init_I] = OS_NO_TASK;    // remove all entries in the prioritylist
    LowestPriority = 0;
#else
    LastExecutedTask = OS_NO_TASK;
#endif

#ifdef OS_IDLE
    OS_Idleprocedure = NULL; // no idle procedure
#endif

    OSRunning      = false; // OS not running yet
    CurrentTask    = OS_NO_TASK;
    LastTask       = OS_NO_TASK;

}

TRegs TmpRegs;
DWORD TmpAddr, TmpStat, TmpSRSCtl, TmpHI, TmpLO;

BYTE NextTask, NextTask_Tmp;
#ifdef OS_PRIORITIES
BYTE Priority, __State, __Priority;
#endif

#ifdef OS_SEMAPHORES
BYTE Check_Events_I;
TTaskData *TaskDataPtr;
TEventData *EventDataPtr;
#endif

#ifdef OS_DELAY_TIMEOUT
BYTE OS_Interrupt_I;
TTaskData *OS_Interrupt_Ptr;
#endif

BYTE OS_Count;
BYTE OS_I;
TTaskData *Timer1Int_Ptr;
DWORD TmpEnInt;

#ifdef OS_SIGNAL_ISR
void Check_Events_ISR() {
    // Transfer events from interrupts to "normal" events
    if (LastEvent != OS_NO_EVENT) {
        for (Check_Events_I; Check_Events_I <= LastEvent; Check_Events_I++) {
            EventDataPtr = &EventData[Check_Events_I];

            // disable interrupts, EventDataPtr^._SemISR can change under higher level interrupt
            InterruptState = DisableInterrupts();
            // ---

            if (EventDataPtr->_SemISR > 0) { // there was signaling from an ISR
                if (EventDataPtr->_Type == et_BINARY)
                EventDataPtr->_SemCount = 1;                                                // binary semaphore
                else
                EventDataPtr->_SemCount = EventDataPtr->_SemCount + EventDataPtr->_SemISR; // counting semaphore
                EventDataPtr->_SemISR = 0;
            }

            // enable interrupts again
            //RestoreInterrupts(InterruptState);
            R30 = InterruptState & 1;
            if (R30 > 0)
            asm EI R0;
            else
            asm DI R0;
            // ---

        }
    }
}
#endif

// ---------------------------
// Timer 1 interrupt procedure
// ---------------------------
void Timer1Int() iv IVT_TIMER_1 ilevel 6 ics ICS_OFF { // <--- iLevel 6 and ICS_OFF
    // -----------------------------------------------------------------------------
    //                                  IMPORANT
    //
    // Make sure no routine calls are done in this interrupt procedure.
    // One can check this to be true in the list file: the first machine
    // statement of this routine should not be "ADDIU SP, SP, -xx"
    //
    // If routines are called then remove the call and place their code inline in the interrupt procedure.
    // -----------------------------------------------------------------------------

    asm rdpgpr SP, SP;

    // -----------------------------------------------------------------------------
    //                              GET CONTEXT
    // -----------------------------------------------------------------------------

    TmpRegs[0] = R2;
    TmpRegs[1] = R3;
    TmpRegs[2] = R4;
    TmpRegs[3] = R5;
    TmpRegs[4] = R6;
    TmpRegs[5] = R7;
    TmpRegs[6] = R8;
    TmpRegs[7] = R9;
    TmpRegs[8] = R10;
    TmpRegs[9] = R11;
    TmpRegs[10] = R12;
    TmpRegs[11] = R13;
    TmpRegs[12] = R14;
    TmpRegs[13] = R15;
    TmpRegs[14] = R16;
    TmpRegs[15] = R17;
    TmpRegs[16] = R18;
    TmpRegs[17] = R19;
    TmpRegs[18] = R20;
    TmpRegs[19] = R21;
    TmpRegs[20] = R22;
    TmpRegs[21] = R23;
    TmpRegs[22] = R24;
    TmpRegs[23] = R25;
    TmpRegs[24] = R26;
    TmpRegs[25] = R27;
    TmpRegs[26] = R28;
    TmpRegs[27] = R29;
    TmpRegs[28] = R30;
    TmpRegs[29] = R31;

    T1IF_bit    = 0; // Clear timer interrupt bit

    // get processor status
    asm MFC0        R30, 12, 0;
    TmpStat = R30; // and save it

    // get processor SRSCtl
    asm MFC0        R30, 12, 2;
    TmpSRSCtl = R30; // and save it

    // get value from HI and LO
    asm MFHI R30;
    TmpHI = R30;

    asm MFLO R30;
    TmpLO = R30;

    // get the "old" Task return address
    asm MFC0        R30, 14, 0;
    TmpAddr = R30;

    // set the status to not handle interrupts with interrupt priority <= than the current one
    // and enable interupts again
    asm {
        MFC0        R2,  13, 0     ; get cause
        SRL         R2,  R2, 10    ; isolate the RIPL bits out of it by right rotating 10 times
        MFC0        R30, 12, 0     ; get processor status
        INS         R30, R2, 10, 6 ; insert the RIPL bits in the IPL field in the status
        INS         R30, R0, 1, 4  ; clear bits 1..4
        MTC0        R30, 12, 0     ; this will enable interrupts again
    }


    // -----------------------------------------------------------------------------
    //                       INCREMENT TIMERTICK COUNTER
    // -----------------------------------------------------------------------------

    if (!Immediate) {
#ifndef OS_NO_AUTO
        OS_Count++; // increment the interrupt counter if no forced interrupt
#endif

        // -----------------------------------------------------------------------------
        //                           TIMED EVENT CHECKING
        // -----------------------------------------------------------------------------
#ifdef OS_DELAY_TIMEOUT
        //Check_Timed_Events; // only decrement the timers if no forced interrupt
        if (LastTask != OS_NO_TASK) {
            for (OS_Interrupt_I = 0; OS_Interrupt_I <= LastTask; OS_Interrupt_I++) {
                OS_Interrupt_Ptr = &TaskData[OS_Interrupt_I];

                if ((OS_Interrupt_Ptr->_State == st_DELAYED) || (OS_Interrupt_Ptr->_State == st_WAITING)) {
                    if (OS_Interrupt_Ptr->_Timer > 0) {
                        OS_Interrupt_Ptr->_Timer--;

                        if (OS_Interrupt_Ptr->_Timer == 0) {
                            switch (OS_Interrupt_Ptr->_State) {
                                #ifdef OS_DELAY
                            case st_DELAYED: //
                                OS_Interrupt_Ptr->_State = st_ELIGIBLE;  // timer was a delay timer
                                Immediate = true;
                                break;
                                #endif

                                #ifdef OS_TIMEOUT
                            case st_WAITING: //
                                OS_Interrupt_Ptr->_Misc.mi_TIMEOUT_bit = 1; // timer was a timeout timer
                                OS_Interrupt_Ptr->_State = st_ELIGIBLE;
                                Immediate = true;
                                #endif
                            }
                        }

                    }
                }

            }
        }
#endif
    }

    // -----------------------------------------------------------------------------
    //                                EVENT CHECKING
    // -----------------------------------------------------------------------------
    // check if there are any pending events (signalled semaphores)

#ifdef OS_SEMAPHORES

#ifdef OS_SIGNAL_ISR
    // Transfer events from interrupts to "normal" events
    if (LastEvent != OS_NO_EVENT) {
        for (Check_Events_I = 0; Check_Events_I <= LastEvent; Check_Events_I++) {
            EventDataPtr = &EventData[Check_Events_I];

            // disable interrupts, EventDataPtr^._SemISR can change under higher level interrupt
            InterruptState = DisableInterrupts(); // --> can be used here (inline code)
            // ---

            if (EventDataPtr->_SemISR > 0) { // there was signaling from an ISR
                if (EventDataPtr->_Type == et_BINARY)
                EventDataPtr->_SemCount = 1;                                                // binary semaphore
                else
                EventDataPtr->_SemCount = EventDataPtr->_SemCount + EventDataPtr->_SemISR;  // counting semaphore
                EventDataPtr->_SemISR = 0;
            }

            // enable interrupts again
            //RestoreInterrupts(InterruptState); --> can not be used here (is a call to a procedure)
            R30 = InterruptState & 1;

            if (R30 > 0)
            asm EI R0;
            else
            asm DI R0;
            // ---

        }
    }
#endif

    if (LastTask != OS_NO_TASK) {
        for (Check_Events_I = 0; Check_Events_I <= LastTask; Check_Events_I++) {
            TaskDataPtr = &TaskData[Check_Events_I];
            // check for waiting on semaphores
            if (TaskDataPtr->_State == st_WAITING) {  // task is waiting for an event
                if (EventData[TaskDataPtr->_EventID]._SemCount > 0) {
                    EventData[TaskDataPtr->_EventID]._SemCount--;
                    TaskDataPtr->_State = st_ELIGIBLE;   // task can be executed
                    Immediate = true;
                }
            }
        }
    }
#endif

    // -----------------------------------------------------------------------------
    //                                TASK SWITCHING
    // -----------------------------------------------------------------------------

    // -----------------------------------------------------------------------------
    //  Save the context of the current task
    // -----------------------------------------------------------------------------

    if (Immediate || (OS_Count >= IntTime)) { // time to switch tasks

        OS_Count  = 0; // always restart the 10 timerticks counter
        Immediate = false;

        // save context of "old" (currently running) task
        Timer1Int_Ptr = NULL;
        if (CurrentTask != OS_NO_TASK)
        Timer1Int_Ptr = &TaskData[CurrentTask];
        else if (IdleRunning)
        Timer1Int_Ptr = &IdleData;

        if (Timer1Int_Ptr != NULL) {
            for (OS_I = 0; OS_I < 30; OS_I++)
            Timer1Int_Ptr->Regs[OS_I] = TmpRegs[OS_I];

            Timer1Int_Ptr->HiVal   = TmpHI;
            Timer1Int_Ptr->LoVal   = TmpLO;
            Timer1Int_Ptr->Status_ = TmpStat;
            Timer1Int_Ptr->RetAddr = TmpAddr;
            Timer1Int_Ptr->SRSCtl  = TmpSRSCtl;
        }

        // -----------------------------------------------------------------------------
        //  Select next task
        // -----------------------------------------------------------------------------

        NextTask = OS_NO_TASK;
        NextTask_Tmp = OS_NO_TASK;

#ifdef OS_PRIORITIES
        // make a task priority table
        if (LastTask != OS_NO_TASK) {
            for (Priority = 0; Priority <= LowestPriority; Priority++) {  // check high to low priorities
                if (LastExecutedTask[Priority] == OS_NO_TASK)
                LastExecutedTask[Priority] = LastTask; // point to last task
                NextTask_Tmp = LastExecutedTask[Priority];  // round robin per priority

                do {
                    NextTask_Tmp++;
                    if (NextTask_Tmp > LastTask)
                    NextTask_Tmp = 0;
                    // wrap around
                    __State    = TaskData[NextTask_Tmp]._State;
                    __Priority = TaskData[NextTask_Tmp]._Priority;
                } while (!(((__State == st_ELIGIBLE) && (__Priority == Priority)) ||  // priority Ok and runnable, or
                (NextTask_Tmp == LastExecutedTask[Priority])));                    // all tasks checked

                if ((__State == st_ELIGIBLE) && (__Priority == Priority)) {
                    NextTask = NextTask_Tmp;
                    LastExecutedTask[Priority] = NextTask;
                    break;
                }
            }
        }

#else // no priorities
        if (LastExecutedTask == OS_NO_TASK)
        LastExecutedTask = LastTask;
        NextTask_Tmp = LastExecutedTask;
        do {
            NextTask_Tmp++;
            if (NextTask_Tmp > LastTask)
            NextTask_Tmp = 0;   // wrap around
        } while (!((TaskData[NextTask_Tmp]._State == st_ELIGIBLE) || // task found, or
        (NextTask_Tmp == LastExecutedTask)));                     // all tasks checked

        if (TaskData[NextTask_tmp]._State == st_ELIGIBLE) {
            NextTask = NextTask_Tmp;
            LastExecutedTask = NextTask;
        }
#endif

        // -----------------------------------------------------------------------------
        //  Load the context of the next task to be executed
        // -----------------------------------------------------------------------------

        if (NextTask != OS_NO_TASK) { // task to execute found
            CurrentTask   = NextTask;
            Timer1Int_Ptr = &TaskData[CurrentTask];
            IdleRunning   = false;
        }
        else  {// no task to execute found, execute the idle task
            CurrentTask   = OS_NO_TASK;
            Timer1Int_Ptr = &IdleData;
            IdleRunning   = true;
        }

        // Restore context of new task
        for (OS_I = 0; OS_I < 30; OS_I++)
        TmpRegs[OS_I] = Timer1Int_Ptr->Regs[OS_I];

        TmpHI   = Timer1Int_Ptr->HiVal;
        TmpLO   = Timer1Int_Ptr->LoVal;
        TmpStat = Timer1Int_Ptr->Status_;
        TmpAddr = Timer1Int_Ptr->RetAddr;
        TmpSRSCtl = Timer1Int_Ptr->SRSCtl;
    }

    // -----------------------------------------------------------------------------
    //                           RESTORE CONTEXT
    // -----------------------------------------------------------------------------

    asm {
        DI
        EHB
    }

    // set the (new) Task's return address
    R30 = TmpAddr;
    asm  MTC0        R30, 14, 0;

    // Restore the HI and LO register values of the new task
    R30 = TmpHI;
    asm MTHI R30;

    R30 = TmpLO;
    asm MTLO R30;

    // Restore the SRSCtl register
    R30 = TmpSRSCtl;
    asm MTC0        R30, 12, 2;

    // Restore the coprocessor's status
    R30 = TmpStat;
    asm MTC0        R30, 12, 0;

    // restore registers for the new task
    R2 = TmpRegs[0];
    R3 = TmpRegs[1];
    R4 = TmpRegs[2];
    R5 = TmpRegs[3];
    R6 = TmpRegs[4];
    R7 = TmpRegs[5];
    R8 = TmpRegs[6];
    R9 = TmpRegs[7];
    R10 = TmpRegs[8];
    R11 = TmpRegs[9];
    R12 = TmpRegs[10];
    R13 = TmpRegs[11];
    R14 = TmpRegs[12];
    R15 = TmpRegs[13];
    R16 = TmpRegs[14];
    R17 = TmpRegs[15];
    R18 = TmpRegs[16];
    R19 = TmpRegs[17];
    R20 = TmpRegs[18];
    R21 = TmpRegs[19];
    R22 = TmpRegs[20];
    R23 = TmpRegs[21];
    R24 = TmpRegs[22];
    R25 = TmpRegs[23];
    R26 = TmpRegs[24];
    R27 = TmpRegs[25];
    R28 = TmpRegs[26];
    R29 = TmpRegs[27];
    R30 = TmpRegs[28];
    R31 = TmpRegs[29];

}

DWORD OS_CreateTask_TmpD;
BYTE *Create_Task_MemPtr;

#ifdef OS_PRIORITIES
TOS_Task OS_CreateTask(TOS_TaskProcPtr TaskProc, TOS_Priority Priority)
#else
TOS_Task OS_CreateTask(TOS_TaskProcPtr TaskProc)
#endif
{
    TOS_Task Result = OS_NO_TASK;

    if (LastTask == (OS_TASKS_COUNT - 1))
    return Result; // no free tasks in the list any more

    if (OSRunning) // disable the OS timer interrupt
    OS_Timer_IE_bit = 0;

    //GetMemAlign(Create_Task_MemPtr, OS_Task_StackSize + 4, 4); // get memory for stack
    Create_Task_MemPtr = (BYTE *)Malloc(OS_Task_StackSize + 4);

    if (Create_Task_MemPtr == NULL) {
        if (OSRunning)
        OS_Timer_IE_bit = 1; // task will be "OS_NO_TASK", enable the OS timer interrupt
        return Result;
    }

    if (LastTask == OS_NO_TASK)
    LastTask = 0;
    else
    LastTask++;

    Result = LastTask;

    OS_CreateTask_TmpD = (DWORD)TaskProc;

    TaskData[LastTask]._State = st_STOPPED;
    #ifdef OS_PRIORITIES
    if (Priority >= OS_PRIORITY_COUNT)
    Priority = OS_PRIORITY_COUNT - 1;
    if (Priority > LowestPriority)
    LowestPriority = Priority;
    TaskData[LastTask]._Priority = Priority;
    #endif

    TaskData[LastTask]._Misc = 0;                  // set all bits to zero

    #ifdef OS_STACK_CHECK
    TaskData[LastTask]._StackBase = (DWORD)Create_Task_MemPtr;
    memset(Create_Task_MemPtr, 0, OS_Task_StackSize + 4);
    #endif

    TaskData[LastTask].Regs[27] = (DWORD)Create_Task_MemPtr + OS_Task_StackSize; // Initial stackpointer = Top of stack. (OS_Task_StackSize bytes stack for each task)
    TaskData[LastTask].RetAddr = OS_CreateTask_TmpD;  // store return address

    asm MFHI R30;
    TaskData[LastTask].HiVal = R30;

    asm MFLO R30;
    TaskData[LastTask].LoVal = R30;

    asm MFC0 R30, 12, 0;
    TaskData[LastTask].Status_ = R30 | 1; // set interrupt enable

    asm MFC0 R30, 12, 2;
    TaskData[LastTask].SRSCtl = R30;

    if (OSRunning) // enable the OS timer interrupt
    OS_Timer_IE_bit = 1;

    return Result;
}

void OS_StartTask(TOS_Task TaskID) {
    if (TaskID > LastTask)
    return; // not a valid TaskID

    // disable the OS timer interrupt
    if (OSRunning)
    OS_Timer_IE_bit = 0;

    if (TaskData[TaskID]._State == st_STOPPED) {
        TaskData[TaskID]._State = st_ELIGIBLE;
        if (OSRunning) {
            Immediate = true;
            T1IF_bit = 1;
        }
    }

    // Restore the OS timer interrupt enable
    if (OSRunning)
    OS_Timer_IE_bit = 1;
}

#ifdef OS_DELAY
void OS_StartTask_Delay(TOS_Task TaskID, WORD Ticks) {
    if (TaskID > LastTask)
    return; // not a valid TaskID

    // disable the OS timer interrupt
    if (OSRunning)
    OS_Timer_IE_bit = 0;

    if (TaskData[TaskID]._State == st_STOPPED) {
        TaskData[TaskID]._State = st_DELAYED; // task is st_DELAYED for a while before starting
        TaskData[TaskID]._Timer = Ticks;
    }

    if (OSRunning) {
        if (Ticks < IntTime) {
            Immediate = true;
            T1IF_bit = 1;
        }
    }

    // Restore the OS timer interrupt enable
    if (OSRunning)
    OS_Timer_IE_bit = 1;
}
#endif

void OS_StopTask(TOS_Task TaskID) {
    if (TaskID > LastTask)
    return; // not a valid TaskId

    // disable the OS timer interrupt
    OS_Timer_IE_bit = 0;

    if (TaskData[TaskID]._State >= st_STOPPED)
    TaskData[TaskID]._State = st_STOPPED; // stop the task

    if (TaskID == CurrentTask) {
        Immediate = true;
        T1IF_bit = 1;
    }

    // Restore the OS timer interrupt enable
    OS_Timer_IE_bit = 1;
}

void OS_ReplaceTask(TOS_Task TaskID) {
    if ((TaskID > LastTask) ||      // not a valid TaskID
            (TaskID == CurrentTask) ||      // can not replace a task by itself
            (CurrentTask == OS_NO_TASK))
    return;

    // disable the OS timer interrupt
    OS_Timer_IE_bit = 0;

    TaskData[CurrentTask]._State = st_STOPPED; // stop the current task
    if (TaskData[TaskID]._State == st_STOPPED)
    TaskData[TaskID]._State = st_ELIGIBLE; // start the new task

    Immediate = true;
    T1IF_bit  = 1;

    // Restore the OS timer interrupt enable
    OS_Timer_IE_bit = 1;
}

TOS_Task OS_CurrentTask() {
    // disable the OS timer interrupt
    TOS_Task Result;

    OS_Timer_IE_bit = 0;

    Result = CurrentTask;

    // Restore the OS timer interrupt enable
    OS_Timer_IE_bit = 1;

    return Result;
}

#ifdef OS_SEMAPHORES
TOS_BinarySemaphore OS_CreateBinarySemaphore(BOOL InitialValue) {
    TOS_BinarySemaphore Result = OS_NO_EVENT; // no free event found

    if (LastEvent == (OS_EVENTS_COUNT - 1))
    return Result; // no unused events available any more

    if (LastEvent == OS_NO_EVENT)
    LastEvent = 0;
    else
    LastEvent++;

    Result = LastEvent;

    EventData[LastEvent]._Type = et_BINARY;
    if (InitialValue)
    EventData[LastEvent]._SemCount = 1;
    else
    EventData[LastEvent]._SemCount = 0;
    #ifdef OS_SIGNAL_ISR
    EventData[LastEvent]._SemISR = 0;
    #endif

    return Result;
}

TOS_CountingSemaphore OS_CreateCountingSemaphore(WORD InitialValue) {
    TOS_CountingSemaphore Result = OS_NO_EVENT; // no free event found

    if (LastEvent == (OS_EVENTS_COUNT - 1))
    return Result; // no unused events available any more

    if (LastEvent == OS_NO_EVENT)
    LastEvent = 0;
    else
    LastEvent++;

    Result = LastEvent;

    EventData[LastEvent]._Type = et_COUNTING;
    EventData[LastEvent]._SemCount = InitialValue;
    #ifdef OS_SIGNAL_ISR
    EventData[LastEvent]._SemISR = 0;
    #endif

    return Result;
}
#endif

BOOL OS_TaskRunning(TOS_Task TaskID) {
    // disable the OS timer interrupt
    BOOL Result;

    OS_Timer_IE_bit = 0;

    Result = st_DESTROYED;
    if (TaskID <= LastTask)
    Result = (TaskData[TaskID]._State > st_STOPPED);

    // Restore the OS timer interrupt enable
    OS_Timer_IE_bit = 1;

    return Result;
}


void OS_Yield() {
    if (CurrentTask == OS_NO_TASK)
    return;

    OS_Timer_IE_bit = 0;

    TaskData[CurrentTask]._State = st_ELIGIBLE; // task is no longer st_RUNNING
    Immediate = true;
    T1IF_bit  = 1;

    OS_Timer_IE_bit = 1;
}


#ifdef OS_DELAY
void OS_Delay(WORD Ticks) {
    if (CurrentTask == OS_NO_TASK)
    return;

    OS_Timer_IE_bit = 0;

    TaskData[CurrentTask]._State = st_DELAYED; // task is st_DELAYED for a while
    TaskData[CurrentTask]._Timer = Ticks;
    Immediate = true;
    T1IF_bit  = 1;  // Immediate action required

    OS_Timer_IE_bit = 1;
}
#endif

#ifdef OS_SEMAPHORES
void OS_SignalSemaphore(TOS_Semaphore Event_) {
    if (Event_ != OS_NO_EVENT) {
        OS_Timer_IE_bit = 0;

        if (EventData[Event_]._Type == et_BINARY)
        EventData[Event_]._SemCount = 1;
        else
        EventData[Event_]._SemCount++;

        OS_Timer_IE_bit = 1;
    }
}

void OS_ClearSemaphore(TOS_Semaphore Event_) {
    if (Event_ != OS_NO_EVENT) {
        OS_Timer_IE_bit = 0;

        EventData[Event_]._SemCount = 0;

#ifdef OS_SIGNAL_ISR
        // clear pending signaling from ISR's

        // disable interrupts, EventDataPtr^._SemISR can change under interrupt
        InterruptState = DisableInterrupts();

        EventData[Event_]._SemISR = 0;

        // enable interrupts again
        RestoreInterrupts(InterruptState);
        // ---
#endif

        OS_Timer_IE_bit = 1;
    }
}

#ifdef OS_SIGNAL_ISR
void OS_SignalSemaphore_ISR(TOS_Semaphore Event_) {
    // disable the OS timer interrupt
    OS_Timer_IE_bit = 0;

    if (Event_ != OS_NO_EVENT)
    EventData[Event_]._SemISR++;

    // Restore the OS timer interrupt enable
    OS_Timer_IE_bit = 1;
}
#endif

#ifdef OS_TIMEOUT
void OS_WaitSemaphore(TOS_Semaphore Event_, WORD TimeOut)
#else
void OS_WaitSemaphore(TOS_Semaphore Event_)
#endif
{
    if ((CurrentTask == OS_NO_TASK) ||
            (Event_ == OS_NO_EVENT))
    return; // semaphore does no exist

    // disable the OS timer interrupt
    OS_Timer_IE_bit = 0;

#ifdef OS_TIMEOUT
    TaskData[CurrentTask]._Misc.mi_TIMEOUT_bit = 0; // clear timeout flag
#endif

#ifdef OS_SIGNAL_ISR
    Check_Events_ISR(); // get the events submitted in ISR and copy then to the actual EventData
#endif


    if (EventData[Event_]._SemCount > 0) // task can be continued immediately
    EventData[Event_]._SemCount--;
    else {  // task has to be suspended

        TaskData[CurrentTask]._EventID = Event_;
        TaskData[CurrentTask]._State   = st_WAITING;
        #ifdef OS_TIMEOUT
        TaskData[CurrentTask]._Timer   = TimeOut;
        #endif

        // Immediate interrupt
        Immediate = true;
        T1IF_bit  = 1;
    }

    // Restore the OS timer interrupt enable
    OS_Timer_IE_bit = 1;
}

#ifdef OS_TIMEOUT
BOOL OS_TimeOut() {
    BOOL Result = false;

    if (CurrentTask == OS_NO_TASK)
    return Result;

    // disable the OS timer interrupt
    OS_Timer_IE_bit = 0;

    Result = (TaskData[CurrentTask]._Misc.mi_TIMEOUT_bit == 1); // timeout flag was set
    TaskData[CurrentTask]._Misc.mi_TIMEOUT_bit = 0;          // clear the flag

    // Restore the OS timer interrupt enable
    OS_Timer_IE_bit = 1;

    return Result;
}
#endif

BOOL OS_ReadSemaphore(TOS_Semaphore Event_) {
    BOOL Result = false;
    if (Event_ <= LastEvent) {
        // disable the OS timer interrupt
        OS_Timer_IE_bit = 0;

        #ifdef OS_SIGNAL_ISR
        Check_Events_ISR(); // get the events submitted in ISR and copy then to the actual EventData
        #endif

        Result = (EventData[Event_]._SemCount > 0);

        // Restore the OS timer interrupt enable
        OS_Timer_IE_bit = 1;
    }

    return Result;
}

WORD OS_ReadCountingSemaphore(TOS_Semaphore Event_) {
    WORD Result = 0;
    if (Event_ <= LastEvent) {
        // disable the OS timer interrupt
        OS_Timer_IE_bit = 0;

        #ifdef OS_SIGNAL_ISR
        Check_Events_ISR(); // get the events submitted in ISR and copy then to the actual EventData
        #endif

        Result = EventData[Event_]._SemCount;

        // Restore the OS timer interrupt enable
        OS_Timer_IE_bit = 1;
    }

    return Result;
}

BOOL OS_TrySemaphore(TOS_Semaphore Event_) {
    BOOL Result = false;
    if (Event_ <= LastEvent) {
        // disable the OS timer interrupt
        OS_Timer_IE_bit = 0;

        #ifdef OS_SIGNAL_ISR
        Check_Events_ISR(); // get the events submitted in ISR and copy then to the actual EventData
        #endif

        Result = (EventData[Event_]._SemCount > 0);

        // Restore the OS timer interrupt enable
        OS_Timer_IE_bit = 1;
    }

    return Result;
}
#endif

#ifdef OS_PRIORITIES
void OS_SetPriority(TOS_Task TaskID, TOS_Priority Priority) {
    // disable the OS timer interrupt
    OS_Timer_IE_bit = 0;

    if (Priority >= OS_PRIORITY_COUNT)
    Priority = (OS_PRIORITY_COUNT - 1);

    if (Priority > LowestPriority)
    LowestPriority = Priority;

    if (TaskID <= LastTask) {
        TaskData[TaskID]._Priority = Priority;

        // Immediate interrupt
        Immediate = true;
        T1IF_bit  = 1;
    }

    // Restore the OS timer interrupt enable
    OS_Timer_IE_bit = 1;
}

TOS_Priority OS_Priority(TOS_Task TaskID) {
    TOS_Priority Result = (OS_PRIORITY_COUNT - 1); // lowest priority

    if (TaskID <= LastTask)
    Result = TaskData[TaskID]._Priority;

    return Result;
}
#endif

TOS_State OS_TaskState(TOS_Task TaskID) {
    // disable the OS timer interrupt
    TOS_State Result;

    OS_Timer_IE_bit = 0;

    Result = st_DESTROYED;
    if (TaskID <= LastTask)
    Result = TaskData[TaskID]._State;

    // Restore the OS timer interrupt enable
    OS_Timer_IE_bit = 1;

    return Result;
}

void OS_EnterCriticalSection() {
    // disable the OS timer interrupt
    OS_Timer_IE_bit = 0;        // disable timer interrupt

    T1IP1_bit       = 0;        // set interrupt Priority to zero (interrupts disabled)
    T1IP2_bit       = 0;        //

    // Restore the OS timer interrupt enable
    OS_Timer_IE_bit = 1;
}

void OS_LeaveCriticalSection() {
    // disable the OS timer interrupt
    OS_Timer_IE_bit = 0;

    T1IP1_bit     = 1;        //  set interrupt Priority again to 6
    T1IP2_bit     = 1;        //

    // Restore the OS timer interrupt enable
    OS_Timer_IE_bit = 1;
}

#ifdef OS_STACK_CHECK
WORD CheckStack_I;
BYTE *CheckStackPtr;

BYTE OS_Stack_Usage(TOS_Task TaskID) {
    BYTE Result = 0;

    if (TaskID == OS_NO_TASK)
    return Result;

    // disable the OS timer interrupt
    OS_Timer_IE_bit = 0;

    if (TaskID != OS_NO_TASK) {
        CheckStack_I = 0;
        CheckStackPtr = (BYTE *)TaskData[TaskID]._StackBase;
        while ((*CheckStackPtr == 0) && (CheckStack_I < OS_Task_StackSize)) {
            CheckStackPtr++;
            CheckStack_I++;
        }
        Result = ((DWORD)(OS_Task_StackSize - CheckStack_I) * 100) / OS_Task_StackSize;
    }

    // Restore the OS timer interrupt enable
    OS_Timer_IE_bit = 1;

    return Result;
}
#endif

void IdleRoutine() {
    while (true) {
#ifdef OS_IDLE
        if (OS_IdleProcedure != NULL)
        OS_IdleProcedure(); // call the idle procedure (if any)
#endif
    }
}

BYTE *OS_Run_MemPtr;
WORD OS_St_Size;

void OS_Run() {
#pragma funcall OS_Run IdleRoutine

#ifdef OS_IDLE
    //GetMemAlign(OS_Run_MemPtr, OS_Task_StackSize + 4, 4);   // get stack of "normal" size
    OS_Run_MemPtr = (BYTE *)malloc(OS_Task_StackSize + 4);

    if (OS_Run_MemPtr == NULL) {  // not enough stack memory available, give a small stack (100 bytes)
        OS_St_Size = 100;
        OS_IdleProcedure = NULL; // no stack available: no idle procedure possible
        OS_Run_MemPtr = (BYTE *)(SP - 104); // give some stack for the Idle Task (there always has to be one)
    }
    else {  // enough stack memory could be obtained
        OS_St_Size =  OS_Task_StackSize;
    }
#else // only a small stack is allocated to the idle routine because there is no Idle procedure
    OS_Run_MemPtr = (BYTE *)(SP - 104); // give some stack for the Idle Task (there always has to be one)
    OS_St_Size = 100;
#endif

    // create IDLE task
    IdleData._State = st_STOPPED;

    IdleData._Misc = 0;                                                  // set all bits to zero

    IdleData.Regs[27] = (DWORD)OS_Run_MemPtr + OS_St_Size;        // initial stackpointer = Top of Stack

    IdleData.RetAddr  = (DWORD)(&IdleRoutine);                      // store return address

    asm MFHI R30;
    IdleData.HiVal = R30;

    asm MFLO R30;
    IdleData.LoVal = R30;

    asm MFC0 R30, 12, 0;
    IdleData.Status_ = R30 | 1; // set interrupt enable

    asm MFC0 R30, 12, 2;
    IdleData.SRSCtl = R30;

    IdleRunning = false;
    Immediate   = true; // give the first trigger to select a task (do not wait for the 10 timerticks)
    OSRunning   = true; // OS is running now
    InitTimer1();          // start the timebase, here we go
}